<?php
/*
 * Plugin: ultimania
 * ~~~~~~~~~~~~~~~~~
 * This plugin is made to replace the database Gerymania. Gerymania was sadly
 * the only global record database supporting the Stunts mode. This wouldn't be
 * the problem, if it would be fast and have just that basic features like
 * a ban system for cheaters (in stunt cheating is very easy :/).
 * This is the reason, why this plugin have been coded.
 * 
 * ----------------------------------------------------------------------------------
 * Author:		askuri
 * Version:		See "const ULTI_PLUGIN_VERSION"
 * Date:		21.06.2022
 * Copyright:	2022 Martin Weber
 * Game:		Trackmania Forever (TMF) only
 * ----------------------------------------------------------------------------------
 *
 * LICENSE: This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * ----------------------------------------------------------------------------------
 *
 * Dependencies: none
 * 
 * Manialink-IDs:
 * - ultimania_pbwidget
 * - ultimania_window
 * 
 * 
 * Action-IDs:
 * - 5450101 = Open records window
 * - 5450102 = Close records window
 * - 5450104 = Open detailed record list
 * - 54502xx = Recordinfo
 */

Aseco::registerEvent('onNewChallenge2',				'ulti_onNewChallenge2'); /** @phpstan-ignore-line */
Aseco::registerEvent('onPlayerFinish',				'ulti_onPlayerFinish'); /** @phpstan-ignore-line */
Aseco::registerEvent('onPlayerConnect',				'ulti_onPlayerConnect'); /** @phpstan-ignore-line */
Aseco::registerEvent('onEndRace1',					'ulti_onEndRace1'); /** @phpstan-ignore-line */
Aseco::registerEvent('onEverySecond',				'ulti_onEverySecond'); /** @phpstan-ignore-line */
Aseco::registerEvent('onPlayerManialinkPageAnswer', 'ulti_onMlAnswer'); /** @phpstan-ignore-line */
Aseco::registerEvent('onMenuLoaded', 				'ulti_onMenuLoaded'); /** @phpstan-ignore-line */
Aseco::registerEvent('onUltimaniaRecordsLoadedApi2', 'ulti_onUltimaniaRecordsLoadedApi2'); /** @phpstan-ignore-line */
Aseco::registerEvent('onUltimaniaRecordApi2',       'ulti_onUltimaniaRecordApi2'); /** @phpstan-ignore-line */

Aseco::addChatCommand('ultirankinfo', 'Shows informations about a record', false); /** @phpstan-ignore-line */
Aseco::addChatCommand('ultiwindow', 'Shows informations about the current map', false); /** @phpstan-ignore-line */
Aseco::addChatCommand('ultiupdate', 'Updates Ultimania', false); /** @phpstan-ignore-line */
Aseco::addChatCommand('ultilist', 'Shows all records', false); /** @phpstan-ignore-line */

const ULTI_PLUGIN_VERSION = '2.0.1';
const ULTI_API_VERSION    = '5'; // DO NOT CHANGE!!! You may get strange or outdated results so keep it as is
const ULTI_ID_PREFIX      = '5450'; // Change this if unexpected things happen when clicking something
const ULTI_MIN_PHP        = '5.6.0'; // Minimum required PHP version
const ULTI_MIN_XASECO     = '1.14'; // Minimum required XAseco version

const ULTI_EVENT_RECORDS_LOADED_LEGACY = 'onUltimaniaRecordsLoaded';
const ULTI_EVENT_RECORDS_LOADED_API2   = 'onUltimaniaRecordsLoadedApi2';
const ULTI_EVENT_RECORD_LEGACY         = 'onUltimaniaRecord';
const ULTI_EVENT_RECORD_API2           = 'onUltimaniaRecordApi2';

global $ulti;
global $ultiMainClass;
global $ultiXasecoAdapter;


class Ultimania {
    /** @var UltimaniaRecords */
    private $records;

    /** @var UltimaniaConfig */
    private $config;

    /** @var UltimaniaXasecoAdapter */
    private $xasecoAdapter;

    /** @var UltimaniaClient */
    private $ultiClient;

    /** @var int timestamp of when the next refresh should happen */
    private $nextRefresh;

    /**
     * @param UltimaniaConfig $ultiConfig
     * @param UltimaniaRecords $ultiRecords
     * @param UltimaniaXasecoAdapter $xasecoAdapter
     * @param UltimaniaClient $client
     */
    public function __construct($ultiConfig, $ultiRecords, $xasecoAdapter, $client) {
        $this->config = $ultiConfig;
        $this->records = $ultiRecords;
        $this->xasecoAdapter = $xasecoAdapter;
        $this->ultiClient = $client;
        $this->nextRefresh = time() + $this->config->getRefreshInterval();
    }

    /**************************
     * BEGIN EVENT LISTENERS
     **/

    /**
     * @return void
     */
    public function onStartup() {
        $this->doEnvironmentChecks();
        $this->registerWithThirdpartyPluginsUpToDate();
    }

    /**
     * @param Challenge $map
     * @return void
     */
    public function onNewChallenge($map) {
        $this->ultiClient->registerOrUpdateMap($map);
        $this->refreshRecordsListAndReleaseEvent();
        $this->showPbWidgetToEveryone();
    }

    /**
     * @param Record $finish_item
     * @return void
     */
    public function onPlayerFinish(Record $finish_item) {
        if ($finish_item->score == 0) return;

        $ultiRecord = $this->mapXasecoRecordToUltiRecord($finish_item);
        $ultiRecord->setAddTime(time());

        $improvement = $this->records->saveRecord(
            $ultiRecord,
            $this->xasecoAdapter->getBestReplayForPlayer($finish_item->player)
        );

        if ($improvement == null) {
            return;
        }

        $newRecord = $improvement->getNewRecord();

        $this->displayPlayerFinishChatMessage($finish_item->player, $improvement);

        $this->releaseUltimaniaRecordEvent($newRecord);
    }

    /**
     * @param Player $player
     * @return void
     */
    public function onPlayerConnect(Player $player) {
        $playerFromApi = $this->ultiClient->registerOrUpdatePlayer($this->mapXasecoPlayerToUltiPlayer($player));

        if ($playerFromApi != null && $playerFromApi->isBanned()) {
            $this->showBannedPlayerInfoWindow($player);
        }

        // Remove Dedimania PB widget in bottom-right corner
        if (function_exists('chat_recpanel')) {
            $command = [];
            $command['author'] = $player;
            $command['params'] = 'RightBottomNoDedi';
            chat_recpanel($this->xasecoAdapter->getAsecoObject(), $command);
        }

        // display global PB widget
        $personalBestRecord = $this->records->getRecordByLogin($player->login);
        $this->pbWidgetShow($player, $personalBestRecord);
    }

    /**
     * @return void
     */
    public function onEndRace() {
        $this->mainWindowHideToEveryone();
        $this->pbWidgetHide();
    }

    /**
     * @return void
     */
    public function onEverySecond() {
        if (time() >= $this->nextRefresh) {
            $this->refreshRecordsListAndReleaseEvent();

            $this->nextRefresh = time() + $this->config->getRefreshInterval();
        }
    }

    /**
     * @param int $playerUid
     * @param Player $player
     * @param int $actionId
     * @return void
     */
    public function onMlAnswer($playerUid, $player, $actionId) {
        switch ($actionId) {
            case ULTI_ID_PREFIX . 101:
                // Close Records Eyepiece windows
                $this->xasecoAdapter->releaseOnPlayerManialinkPageAnswerEvent($playerUid, $player, 91800);

                // Close standard windows like /list
                $this->xasecoAdapter->releaseOnPlayerManialinkPageAnswerEvent($playerUid, $player, 0);

                $this->mainWindowShow($player);
                break;

            case ULTI_ID_PREFIX . 102:
                $this->mainWindowHideToPlayer($player);
                break;

            case ULTI_ID_PREFIX . 104:
                $this->showFullRecordList($player);
                break;

            case ULTI_ID_PREFIX . 105:
                $this->ultiClient->toggleAllowReplayDownlod($player->login);

                // refresh the list so the player sees he can't even download his own replay
                $this->refreshRecordsListAndReleaseEvent();

                // refresh the window
                $this->mainWindowShow($player);
                break;

            default: // special cases
                // Prefix for recordinfos is 2
                if (substr((string) $actionId, 0, strlen(ULTI_ID_PREFIX) + 1) == ULTI_ID_PREFIX . 2) {
                    $rank = intval(substr((string) $actionId, strlen(ULTI_ID_PREFIX) + 1)) + 1;

                    $this->showUltiRankInfo($player, $rank);
                    $this->mainWindowHideToPlayer($player);
                }
        }
    }

    /**
     * @param Player $author
     * @return void
     */
    public function onChatUltiList(Player $author) {
        $this->showFullRecordList($author);
    }

    /**
     * @param Player $author
     * @param string $rank starting at 1
     * @return void
     */
    public function onChatUltiRankInfo(Player $author, $rank) {
        if (!is_numeric($rank)) {
            $this->xasecoAdapter->chatSendServerMessageToPlayer('$ff0> $f00Usage: $bbb/ultirankinfo ranknumber$f00  (e.g. $bbb/ultirankinfo 1$f00)', $author);
            return;
        }

        $this->showUltiRankInfo($author, (int) $rank);
    }

    /**
     * @param Player $author
     * @return void
     */
    public function onChatUltiWindow(Player $author) {
        $this->mainWindowShow($author);
    }

    /**
     * @param Player $author
     * @return void
     */
    public function onChatUltiUpdate(Player $author) {
        if ($this->xasecoAdapter->isMasterAdmin($author)) {
            $newestVersion = $this->ultiClient->fetchNewestAvailableUltimaniaClientVersion();

            if ($newestVersion !== false) {
                /** @var String $newestVersion */
                if (version_compare(ULTI_PLUGIN_VERSION, $newestVersion) == -1) {
                    $content = $this->ultiClient->fetchUpdate($newestVersion);

                    if ($content) {
                        if (!file_put_contents(getcwd() . '/plugins/plugin.ultimania.php', $content)) {
                            $this->xasecoAdapter->chatSendServerMessageToPlayer('$f00Unable to replace "plugin.ultimania.php". Please check the file permissions and try again.', $author);
                        } else {
                            $this->xasecoAdapter->chatSendServerMessageToPlayer('$0f0Successfully updated. Please restart XAseco.', $author);
                        }
                    } else {
                        $this->xasecoAdapter->chatSendServerMessageToPlayer('$f00Getting newest version file failed!', $author);
                    }
                } else {
                    $this->xasecoAdapter->chatSendServerMessageToPlayer('$0f0You are using the newest version.', $author);
                }
            } else {
                $this->xasecoAdapter->chatSendServerMessageToPlayer('$f00Unable to get current version information!', $author);
            }
        } else {
            $this->xasecoAdapter->chatSendServerMessageToPlayer('$f00No permissions!', $author);
        }
    }

    /**
     * END EVENT LISTENERS
     **************************/

    /*************************************
     * BEGIN MANIALINKS
     */

    /**
     * @return void
     */
    private function mainWindowShow(Player $player) {
        $ultiInfoRaw = $this->ultiClient->fetchInfotextInMainWindow();
        if ($ultiInfoRaw === false) {
            $ultinfo = '';
        } else {
            $ultinfo = htmlspecialchars($ultiInfoRaw);
        }

        $xml = '<manialink id="ultimania_window">
			
			<quad posn="0 3 0" sizen="79 54" bgcolor="001a" halign="center" valign="center" />
			<quad posn="0 3 2" sizen="81 56" style="Bgs1InRace" substyle="BgCard3" halign="center" valign="center" />
			
			<quad posn="0 29.5 1" sizen="82 5" style="BgsPlayerCard" substyle="BgRacePlayerLine" halign="center" />
			<quad posn="0 30 0.5" sizen="80 5" style="Bgs1InRace" substyle="BgTitle3_3" halign="center" />
			<label posn="0 29 2" style="TextRankingsBig" halign="center" text="Ultimania World Records" />
			
			<quad posn="33.5 30 3" sizen="5 5" style="Icons128x128_1" substyle="BackFocusable" action="5450102" />
			
			<label posn="-38 -23.65 3" text="$000$tUltimania TMF v' . ULTI_PLUGIN_VERSION . ' using API ' . ULTI_API_VERSION . '" textsize="0" />
		';

        $xml .= '
			<quad posn="-38 1 1" sizen="37 47" style="BgsPlayerCard" substyle="BgRacePlayerName" valign="center" />
			<quad posn="-0.5 24.5 1" sizen="38.5 41.5" style="BgsPlayerCard" substyle="BgRacePlayerName" />
			<quad posn="-0.5 -17.5 1" sizen="38.5 5" style="BgsPlayerCard" substyle="BgRacePlayerName" />
			
			<label posn="0 24 2" text="' . $ultinfo . '" textsize="2" />
		';

        $xml .= '<frame posn="-38.5 26.5 1">';

        if ($this->records->isEmpty()) {
            $xml .= '<label posn="1.8 -3 1" text="$oCurrently no records" textsize="2" />';
        } else {
            $x = 0;
            $y = -3;

            /**
             * @var int $i rank - 1
             * @var UltimaniaRecord $record
             */
            foreach ($this->records->getLimitedBy($this->config->getNumberOfRecordsDisplayLimit()) as $i => $record) {
                $rank = $i + 1;
                switch ($rank) {
                    case 1:
                        $xml .= '<quad posn="' . ($x + 1.5) . ' ' . $y . ' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="First" />';
                        break;
                    case 2:
                        $xml .= '<quad posn="' . ($x + 1.5) . ' ' . $y . ' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Second" />';
                        break;
                    case 3:
                        $xml .= '<quad posn="' . ($x + 1.5) . ' ' . $y . ' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Third" />';
                        break;

                    default:
                        $xml .= '<label posn="' . ($x + 1.2) . ' ' . $y . ' 1" text="' . $rank . '." textsize="1" />';
                }

                $xml .= '<label posn="' . ($x + 4) . ' ' . $y . ' 1" text="' . $record->getScore() . '" textsize="1" />';
                $xml .= '<label posn="' . ($x + 8) . ' ' . $y . ' 1" text="' . $this->handleSpecialChars($record->getPlayer()->getNick()) . '" sizen="12 2" textsize="1" />';
                $xml .= '<label posn="' . ($x + 21) . ' ' . $y . ' 1" text="' . $record->getPlayer()->getLogin() . '" sizen="13 2" textsize="1" />';

                if ($record->isReplayAvailable()) {
                    $xml .= '<quad posn="' . ($x + 34.2) . ' ' . ($y + 0.2) . ' 2" sizen="1.8 1.8" style="Icons128x128_1" substyle="Replay" manialink="'.$this->ultiClient->getLinkForViewReplayManialink($record).'" />';
                }
                $xml .= '<quad posn="' . ($x + 35.8) . ' ' . ($y - 0.1) . ' 2" sizen="1.35 1.35" style="Icons64x64_1" substyle="TrackInfo" action="' . ULTI_ID_PREFIX . 2 . $i . '" />';

                if ($record->getPlayer()->getLogin() == $player->login) {
                    $xml .= '<quad posn="' . ($x + 0.7) . ' ' . ($y + 0.2) . ' 0.5" sizen="36.65 2" style="BgsPlayerCard" substyle="ProgressBar" />';
                }

                $y -= 1.83;
            }
        }
        $xml .= '</frame>';

        $ultimaniaPlayer = $this->ultiClient->getPlayerInfo($player->login);
        if ($ultimaniaPlayer != null) {
            $allowReplayDownloadSubstyle = $ultimaniaPlayer->isAllowReplayDownload() == true ? "LvlGreen" : "LvlRed";
            $xml .= '<frame posn="0.2 -20 0">';
            $xml .= '<quad posn="0.4 0 5" sizen="1.5 1.5" style="Icons64x64_1" substyle="' . $allowReplayDownloadSubstyle . '" valign="center"/>';
            $xml .= '<label posn="0 0 1" style="CardButtonSmallWide" text="Allow players to view my replays" valign="center" scale="0.77" action="' . ULTI_ID_PREFIX . 105 . '"/>';
            $xml .= '</frame>';
        }

        $xml .= '</manialink>';

        $this->xasecoAdapter->sendManialinkToPlayer($player, $xml);
    }

    /**
     * @param Player $player
     * @return void
     */
    private function mainWindowHideToPlayer($player) {
        $xml = '<manialink id="ultimania_window"></manialink>';
        $this->xasecoAdapter->sendManialinkToPlayer($player, $xml);
    }

    /**
     * @return void
     */
    private function mainWindowHideToEveryone() {
        $xml = '<manialink id="ultimania_window"></manialink>';
        $this->xasecoAdapter->sendManialinkToEveryone($xml);
    }

    /**
     * Record may be null, so we need the Player object to show a "no record yet" widget to the player
     * @param Player $player
     * @param UltimaniaRecord|null $record
     * @return void
     */
    private function pbWidgetShow($player, $record) {
        if (empty($record)) {
            $pbText = '---';
        } else {
            $pbText = $record->getScore();
        }

        $xml = '<manialink id="ultimania_pbwidget">
			<frame posn="53.5 -32.7 0">
                <label posn="4.5 -6.2 2" sizen="10 2" halign="right" valign="center" textsize="1" textcolor="dddf" text="Ultimania PB:"/>
                <label posn="5.0 -6.2 2" sizen="6.5 2" halign="left" valign="center" textsize="1" textcolor="ffff" text="  ' . $pbText . '" action="' . ULTI_ID_PREFIX . '104"/>
			</frame>
		</manialink>';

        $this->xasecoAdapter->sendManialinkToPlayer($player, $xml);
    }

    /**
     * @return void
     */
    private function pbWidgetHide() {
        $xml = '<manialink id="ultimania_pbwidget"></manialink>';
        $this->xasecoAdapter->sendManialinkToEveryone($xml);
    }

    /**
     * @param Player $player
     * @return void
     */
    private function showFullRecordList(Player $player) {
        if ($this->records->isEmpty()) {
            $this->xasecoAdapter->chatSendServerMessageToPlayer('$06fNo Records. Feel free to drive the first :)', $player);
            return;
        }

        $header = 'Ultimania records:'; // Window Header

        $player->msgs = [];

        // Settings
        $recordsPerPage = 15;
        $rankOfPlayer = $this->records->getRankOfPlayer($player->login);
        $startpage = $rankOfPlayer === false ? 1 : ceil($rankOfPlayer / $recordsPerPage);
        $player->msgs[0] = [$startpage,
            $header,
            array(1.4, 0.1, 0.15, 0.3, 0.35, 0.3, 0.2), // widths: overall, col1, col2, col3, ...
            array('Icons64x64_1', 'TrackInfo') // icon style
        ];

        $page = 1;
        foreach ($this->records->getAll() as $id => $record) {
            $time = $this->timestampToDateTimeStringOrGery($record->getAddTime());
            $playerHighlight = $record->getPlayer()->getLogin() == $player->login ? '{#logina}' : '';
            $player->msgs[$page][] = [
                $id + 1,
                $record->getScore(),
                $record->getPlayer()->getNick(),
                $playerHighlight . $record->getPlayer()->getLogin(),
                $time,
                $record->isReplayAvailable() ? '$h['.$this->ultiClient->getLinkForViewReplayManialink($record).']Replay' : '',
            ];

            if ((($id + 1) % $recordsPerPage) == 0) $page++;
        }

        // display ManiaLink message
        display_manialink_multi($player);
    }

    /**
     * @param Player $showToPlayer player to show the information to
     * @param int $rank starting from 1. Rank that information should be shown about
     * @return void
     */
    private function showUltiRankInfo(Player $showToPlayer, $rank) {
        $record = $this->records->getRecordByRank($rank);
        if ($record === null) {
            $this->xasecoAdapter->chatSendServerMessageToPlayer('$ff0>> $f00This record does not exists', $showToPlayer);
            return;
        }

        $dateTimeStringOrGery = $this->timestampToDateTimeStringOrGery($record->getAddTime());

        $header = 'Ultimania record information:';

        $showToPlayer->msgs = array();
        $showToPlayer->msgs[0] = array(1, // startpage
            $header,
            array(0.8, 0.3, 0.5), // widths: overall, col1, col2, col3, ...
            array('Icons64x64_1', 'TrackInfo') // icon style
        );

        $showToPlayer->msgs[] = array(array('Score', '$FFC' . $record->getScore() . ' Points'),
            array('Login', '$FFC' . $record->getPlayer()->getLogin()),
            array('Nickname', '$FFC' . $record->getPlayer()->getNick()),
            array('Recorded on', '$FFC' . $dateTimeStringOrGery),
            array('$h[' . $this->ultiClient->getLinkForAdminRecManialink($record->getPlayer(), $this->xasecoAdapter->getCurrentChallengeObject()->uid) . ']Admin')
        );

        // display ManiaLink message
        display_manialink_multi($showToPlayer);
    }

    /**
     * @param Player $player
     * @return void
     */
    private function showBannedPlayerInfoWindow(Player $player) {
        $header = 'Ultimania global record database information:';
        $data = array();
        $data[] = array('You\'re banned from the global records database Ultimania.');
        $data[] = array('');
        $data[] = array('That means:');
        $data[] = array('- Your record won\'t be saved on Ultimania');
        $data[] = array('- You will always see this window when joining a server');
        $data[] = array('');
        $data[] = array('You can be unbanned by writing an apology for whatever you\'ve done');
        $data[] = array('(cheating, hacking, ...) to askuri (enwi2@t-online.de)');
        $data[] = array('If he thinks you regret whatever you did, he\'ll unban you.');
        $data[] = array('');
        $data[] = array('Best regards');
        display_manialink($player->login, $header, array('Icons64x64_1', 'TrackInfo', -0.01), $data, array(1.1), 'OK');
    }

    /**
     * END MANIALINKS
     *************************************/

    /**********************
     * BEGIN HELPER METHODS
     */

    // stolen from records_eyepiece
    /**
     * @param string $string
     * @return string
     */
    private function handleSpecialChars($string) {
        // Remove links, e.g. "$(L|H|P)[...]...$(L|H|P)"
        $string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)\$(L|H|P)/i', '$2', $string);
        $string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)/i', '$2', $string);
        $string = preg_replace('/\${1}(L|H|P)(.*?)/i', '$2', $string);

        // Remove formatting $S $H $W $I $L $O $N
        $string = preg_replace('/\${1}[SHWILON]/i', '', $string);

        // Convert & " ' > <

        $string = str_replace(array('&', '"', "'", '>', '<'),
            array('&amp;', '&quot;', '&apos;', '&gt;', '&lt;'),
            $string
        );
        $string = stripNewlines($string);    // stripNewlines() from basic.inc.php

        return validateUTF8String($string);    // validateUTF8String() from basic.inc.php
    }

    /**
     * @return void
     */
    private function doEnvironmentChecks() {
        if (version_compare(phpversion(), ULTI_MIN_PHP, '<')) {
            trigger_error('[Ultimania] Not supported PHP version (' . phpversion() . ')! Please update to min. version ' . ULTI_MIN_PHP . '!', E_USER_ERROR);
        }

        if (!function_exists('curl_init')) {
            trigger_error('[Ultimania] You need to enable cURL in php.ini!', E_USER_ERROR);
        }

        // Check for the right XAseco-Version
        if (defined('XASECO_VERSION')) {
            if (version_compare(XASECO_VERSION, ULTI_MIN_XASECO, '<')) { /* @phpstan-ignore-line PHPStan takes the XASECO_VERSION from the blank xaseco installation in this project, which is always fulfilling the condition */
                trigger_error('[Ultimania] Not supported XAseco version (' . XASECO_VERSION . ')! Please update to min. version ' . ULTI_MIN_XASECO . '!', E_USER_ERROR);
            }
        } else {
            trigger_error('[Ultimania] Can not identify the System, "XASECO_VERSION" is unset! This plugin runs only with XAseco/' . ULTI_MIN_XASECO . '+', E_USER_ERROR);
        }

        if ($this->xasecoAdapter->getGameType() != 'TMF') {
            trigger_error('[Ultimania] This plugin supports TMF only. Can not start with a "' . $this->xasecoAdapter->getGameType() . '" Dedicated-Server!', E_USER_ERROR);
        }

        if ($this->xasecoAdapter->getGameInfo()->mode != Gameinfo::STNT) {
            trigger_error('[Ultimania] This plugin supports stunts game mode only. Current game mode is: ' . $this->xasecoAdapter->getGameInfo()->getMode(), E_USER_ERROR);
        }
    }

    /**
     * @param int $timestamp
     * @return string
     */
    private function timestampToDateTimeStringOrGery($timestamp) {
        // Geryimports were done on 2013/24/03 and other dates. I keep searching them
        if ($timestamp > 1364148372 and $timestamp <= 1367798399) {
            $time = 'Gerymania Import';
        } else {
            $time = date('M j Y G:i', $timestamp);
        }

        return $time;
    }

    /**
     * Register this to the global version pool (for up-to-date checks)
     * @return void
     */
    private function registerWithThirdpartyPluginsUpToDate() {
        $this->xasecoAdapter->registerWithThirdpartyPluginsUpToDate(
            'plugin.ultimania.php',
            'askuri',
            ULTI_PLUGIN_VERSION
        );
    }

    /**
     * @return void
     */
    private function releaseRecordsLoadedEvent() {
        $this->xasecoAdapter->releaseEvent(ULTI_EVENT_RECORDS_LOADED_API2, $this->records->getAll());
    }

    /**
     * @param UltimaniaRecord $record
     * @return void
     */
    private function releaseUltimaniaRecordEvent(UltimaniaRecord $record) {
        $this->xasecoAdapter->releaseEvent(ULTI_EVENT_RECORD_API2, $record);
    }

    /**
     * @return void
     */
    private function refreshRecordsListAndReleaseEvent() {
        $this->records->refresh($this->xasecoAdapter->getCurrentChallengeObject()->uid);
        $this->releaseRecordsLoadedEvent();
    }

    /**
     * @param Player $player Player who drove the record
     * @param UltimaniaRecordImprovement $improvement
     * @return void
     */
    private function displayPlayerFinishChatMessage(Player $player, UltimaniaRecordImprovement $improvement) {
        $message = $this->generatePlayerFinishMessage($player, $improvement);

        if ($message) {
            if ($improvement->getNewRank() <= $this->config->getNumberOfRecordsDisplayLimit()) {
                // only show the message publicly, if it's within the display limit
                $this->xasecoAdapter->chatSendServerMessage($this->xasecoAdapter->formatColors($message));
            } else {
                // if the score is not that good, show it to the player though
                $message = str_replace('{#server}>> ', '{#server}> ', $message);
                $this->xasecoAdapter->chatSendServerMessageToPlayer($this->xasecoAdapter->formatColors($message), $player);
            }
        }
    }

    /**
     * Public only for testing. Would be too much work to actually fix the architectural issue behind that.
     * @param Player $player Player who drove the record
     * @param UltimaniaRecordImprovement $improvement
     * @return string|null message or null if unknown record type
     */
    public function generatePlayerFinishMessage($player, UltimaniaRecordImprovement $improvement) {
        $nicknameWithoutColors = stripColors($player->nickname);
        switch ($improvement->getType()) {
            case UltimaniaRecordImprovement::TYPE_NEW:
                return formatText($this->config->getMessageRecordNew(),
                    $nicknameWithoutColors,
                    $improvement->getNewRank(),
                    $improvement->getNewRecord()->getScore(),
                    $improvement->getPreviousRank(),
                    $improvement->getRecordDifference()
                );
            case UltimaniaRecordImprovement::TYPE_EQUAL:
                return formatText($this->config->getMessageRecordEqual(),
                    $nicknameWithoutColors,
                    $improvement->getNewRank(),
                    $improvement->getNewRecord()->getScore()
                );
            case UltimaniaRecordImprovement::TYPE_NEW_RANK:
                return formatText($this->config->getMessageRecordNewRank(),
                    $nicknameWithoutColors,
                    $improvement->getNewRank(),
                    $improvement->getNewRecord()->getScore(),
                    $improvement->getPreviousRank(),
                    $improvement->getRecordDifference()
                );
            case UltimaniaRecordImprovement::TYPE_FIRST:
                return formatText($this->config->getMessageRecordFirst(),
                    $nicknameWithoutColors,
                    $improvement->getNewRank(),
                    $improvement->getNewRecord()->getScore()
                );
            default:
                return null;
        }
    }

    /**
     * @param Record $xasecoRecord
     * @return UltimaniaRecord
     */
    private function mapXasecoRecordToUltiRecord(Record $xasecoRecord) {
        return new UltimaniaRecord(
            $this->mapXasecoPlayerToUltiPlayer($xasecoRecord->player),
            $xasecoRecord->challenge->uid,
            $xasecoRecord->score
        );
    }

    /**
     * @param Player $xasecoPlayer
     * @return UltimaniaPlayer
     */
    private function mapXasecoPlayerToUltiPlayer(Player $xasecoPlayer) {
        return new UltimaniaPlayer($xasecoPlayer->login, $xasecoPlayer->nickname);
    }

    /**
     * @return void
     */
    private function showPbWidgetToEveryone() {
        $recordsByLogin = $this->records->getRecordsIndexedByLogin();
        foreach ($this->xasecoAdapter->getPlayerList() as $p) {
            $this->pbWidgetShow($p, $recordsByLogin[$p->login]);
        }
    }

    /**
     * END HELPER METHODS
     ********************/
}

class UltimaniaConfig {
    // non-configurable via XML
    /** @var int */
    private $refresh_interval = 180;
    /** @var int */
    private $connect_timeout = 10;
    /** @var int */
    private $request_timeout = 10;
    /** @var int */
    private $number_of_records_display_limit = 25;

    // configurable via XML with defaults in this file
    /** @var string|null */
    private $messageRecordNew;
    /** @var string|null */
    private $messageRecordEqual;
    /** @var string|null */
    private $messageRecordNewRank;
    /** @var string|null */
    private $messageRecordFirst;

    /**
     * Creates an object of this calls and populates it with the values from the given xml file.
     * It's not necessary to instantiate this class with a config file. Just create it with the
     * usual constructor and it will take the defaults.
     *
     * @param string $filename
     * @return UltimaniaConfig
     */
    public static function instantiateFromFile($filename) {
        $ultiConfig = new self();

        if (file_exists($filename)) {
            $rawConfig = simplexml_load_file($filename);
            $ultiConfig->messageRecordNew = (string)$rawConfig->messages->record_new; /* @phpstan-ignore-line */
            $ultiConfig->messageRecordEqual = (string)$rawConfig->messages->record_equal; /* @phpstan-ignore-line */
            $ultiConfig->messageRecordNewRank = (string)$rawConfig->messages->record_new_rank; /* @phpstan-ignore-line */
            $ultiConfig->messageRecordFirst = (string)$rawConfig->messages->record_first; /* @phpstan-ignore-line */
        }

        return $ultiConfig;
    }

    /**
     * @return int
     */
    public function getRefreshInterval() {
        return $this->refresh_interval;
    }

    /**
     * @return int
     */
    public function getConnectTimeout() {
        return $this->connect_timeout;
    }

    /**
     * @return int
     */
    public function getRequestTimeout() {
        return $this->request_timeout;
    }

    /**
     * @return int
     */
    public function getNumberOfRecordsDisplayLimit() {
        return $this->number_of_records_display_limit;
    }

    /**
     * @return string
     */
    public function getMessageRecordNew() {
        return !empty($this->messageRecordNew) ? $this->messageRecordNew
            : '{#server}>> {#highlite}{1}{#record} secured his/her {#rank}{2}{#record}. Ultimania Record! Score: {#highlite}{3}{#record} $n({#rank}{4}{#highlite}+{5}{#record})';
    }

    /**
     * @return string
     */
    public function getMessageRecordEqual() {
        return !empty($this->messageRecordEqual) ? $this->messageRecordEqual
            : '{#server}>> {#highlite}{1}{#record} equaled his/her {#rank}{2}{#record}. Ultimania Record! Score: {#highlite}{3}';
    }

    /**
     * @return string
     */
    function getMessageRecordNewRank() {
        return !empty($this->messageRecordNewRank) ? $this->messageRecordNewRank
            : '{#server}>> {#highlite}{1}{#record} gained the {#rank}{2}{#record}. Ultimania Record! Score: {#highlite}{3}{#record} $n({#rank}{4}{#highlite}+{5}{#record})';
    }

    /**
     * @return string
     */
    public function getMessageRecordFirst() {
        return !empty($this->messageRecordFirst) ? $this->messageRecordFirst
            : '{#server}>> {#highlite}{1}{#record} claimed the {#rank}{2}{#record}. Ultimania Record! Score: {#highlite}{3}';
    }
}

class UltimaniaDtoMapper {

    /**
     * @param array{'player': array{"login": string, "nick": "string"}, "nick": string, "score": int, "updated_at": int}[]|null $dtos
     * @return UltimaniaRecord[]
     */
    public function mapRecordDtosToUltiRecords($dtos) {
        if (empty($dtos)) {
            return [];
        }
        return array_map('self::mapRecordDtoWithPlayerDtoToUltiRecord', $dtos); /* @phpstan-ignore-line PHPStand doesn't understand this string is a callback. */
    }

    /**
     * @param array{'player': array{"login": string, "nick": "string", "allow_replay_download": bool, "banned": bool}, "id": string|null, "map_uid": string, "score": int, "updated_at": int, "replay_available": bool} $dto
     * @return UltimaniaRecord
     */
    public function mapRecordDtoWithPlayerDtoToUltiRecord($dto) {
        return new UltimaniaRecord(
            $this->mapPlayerDtoToUltiPlayer($dto['player']),
            $dto['map_uid'],
            $dto['score'],
            $dto['updated_at'],
            $dto['id'],
            $dto['replay_available']
        );
    }

    /**
     * @param array{'player_login': string, 'map_uid': string, "id": string|null, "score": int, "updated_at": int, "replay_available": bool} $dto
     * @return UltimaniaRecord
     */
    public function mapRecordDtoToUltiRecord($dto) {
        return new UltimaniaRecord(
            $dto['player_login'],
            $dto['map_uid'],
            $dto['score'],
            $dto['updated_at'],
            $dto['id'],
            $dto['replay_available']
        );
    }

    /**
     * @param UltimaniaRecord $ultimaniaRecord
     * @param string $mapUid
     * @return array{'player_login': string, "map_uid": string, "score": int}
     */
    public function mapUltiRecordToRecordDto(UltimaniaRecord $ultimaniaRecord, $mapUid) {
        return [
            'player_login' => $ultimaniaRecord->getPlayer()->getLogin(),
            'map_uid' => $mapUid,
            'score' => $ultimaniaRecord->getScore()
        ];
    }

    /**
     * @param Challenge $map
     * @return array{'uid': string, 'name': string}
     */
    public function mapXasecoChallengeToMapDto(Challenge $map) {
        return [
            'uid' => $map->uid,
            'name' => $map->name,
        ];
    }

    /**
     * @param array{'login': string, 'nick': string, 'allow_replay_download': bool, 'banned': bool}|null $playerDto
     * @return UltimaniaPlayer|null
     */
    public function mapPlayerDtoToUltiPlayer($playerDto) {
        if (empty($playerDto)) {
            return null;
        }
        return new UltimaniaPlayer(
            $playerDto['login'],
            $playerDto['nick'],
            $playerDto['allow_replay_download'],
            $playerDto['banned']
        );
    }

    /**
     * @param UltimaniaPlayer $player
     * @return array{'login': string, 'nick': string, 'is_allow_replay_download'?: bool}
     */
    public function mapUltiPlayerToPlayerDto(UltimaniaPlayer $player) {
        $dto = [
            'login' => $player->getLogin(),
            'nick' => $player->getNick(),
        ];
        if ($player->isAllowReplayDownload() !== null) {
            $dto['allow_replay_download'] = $player->isAllowReplayDownload();
        }

        return $dto;
    }
}

class UltimaniaClient {

    /** @var UltimaniaConfig  */
    private $config;

    /** @var UltimaniaDtoMapper */
    private $dtoMapper;

    /** @var string */
    private $ultimaniaApiServerUrl = 'http://ultimania5.askuri.de';

    /** @var string */
    private $adminRecManialinkUrl = 'ulti:admin_rec';

    /** @var string */
    private $viewReplayManialinkUrl = 'ulti:view_replay';

    /**
     * @param UltimaniaConfig $config
     * @param UltimaniaDtoMapper $dtoMapper
     */
    public function __construct($config, $dtoMapper) {
        $this->config = $config;
        $this->dtoMapper = $dtoMapper;
    }

    /**
     * @param string $mapUid
     * @param int|null $limit
     * @return UltimaniaRecord[]
     */
    public function getRecords($mapUid, $limit = null) {
        return $this->dtoMapper->mapRecordDtosToUltiRecords(
            $this->doRequest('GET',
                'maps/' . $mapUid . '/records',
                ['limit' => $limit]
            )['response']
        );
    }

    /**
     * @param string $recordId
     * @param string $replayContent
     * @return array{"replay_available": bool}
     */
    public function submitReplay($recordId, $replayContent) {
        return $this->doRequest(
            'POST',
            'records/' . $recordId . '/replay',
            $replayContent,
            true
        )['response'];
    }

    /**
     * @param UltimaniaRecord $record
     * @param string $mapUid
     * @return UltimaniaRecord|null return the record including its assigned ID or null if unsuccessful (e.g. player banned)
     */
    public function submitRecord($record, $mapUid) {
        $response = $this->doRequest(
            'PUT',
            'records',
            $this->dtoMapper->mapUltiRecordToRecordDto($record, $mapUid)
        )['response'];

        return $response == null ? null : $this->dtoMapper->mapRecordDtoToUltiRecord($response);
    }

    /**
     * @param string $playerLogin
     * @return UltimaniaPlayer|null
     */
    public function getPlayerInfo($playerLogin) {
        return $this->dtoMapper->mapPlayerDtoToUltiPlayer(
            $this->doRequest(
                'GET',
                'players/' . $playerLogin,
                []
            )['response']
        );
    }

    /**
     * @param UltimaniaPlayer $player
     * @return UltimaniaPlayer|null
     */
    public function registerOrUpdatePlayer($player) {
        return $this->dtoMapper->mapPlayerDtoToUltiPlayer(
            $this->doRequest(
                'PUT',
                'players',
                $this->dtoMapper->mapUltiPlayerToPlayerDto($player)
            )['response']
        );
    }

    /**
     * @param string $playerLogin
     * @return void
     */
    public function toggleAllowReplayDownlod($playerLogin) {
        $ultiPlayer = $this->getPlayerInfo($playerLogin);
        if ($ultiPlayer) {
            $ultiPlayer->setAllowReplayDownload(!$ultiPlayer->isAllowReplayDownload());
            $this->registerOrUpdatePlayer($ultiPlayer);
        }
    }

    /**
     * @param Challenge $map
     * @return void
     */
    public function registerOrUpdateMap($map) {
        $this->doRequest(
            'PUT',
            'maps',
            $this->dtoMapper->mapXasecoChallengeToMapDto($map)
        );
    }

    /**
     * @param string $method "GET"|"POST"|"PUT"
     * @param string $endpoint Path in URL after /api/v5/ (no leading slashes needed in this parameter)
     * @param mixed $payload When GET, these are query parameters, otherwise these are put as JSON in the body
     * @param bool $isBinaryRequest If it's binary, it won't be json_encoded. Instead, it goes straight to the body.
     * @return array{"response": mixed|null, "httpcode": int}
     */
    private function doRequest($method, $endpoint, $payload, $isBinaryRequest = false) {
        $handle = curl_init();

        $url = $this->ultimaniaApiServerUrl . '/api/v' . ULTI_API_VERSION . '/' . $endpoint;
        $requestId = uniqid();
        $httpHeaders = ['Accept: application/json', 'X-Request-Id: ' . $requestId];

        curl_setopt($handle, CURLOPT_USERAGENT, 'TMF\Xaseco' . XASECO_VERSION . '\Ultimania' . ULTI_PLUGIN_VERSION . '\API' . ULTI_API_VERSION);
        curl_setopt($handle, CURLOPT_HEADER, false);
        curl_setopt($handle, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($handle, CURLOPT_CUSTOMREQUEST, $method);
        if ($method == 'GET') {
            $url .= http_build_query($payload);
        } else {
            if ($isBinaryRequest) {
                $httpHeaders[] = 'Content-Type: application/octet-stream';
            } else {
                $payload = json_encode($payload);
                $httpHeaders[] = 'Content-Type: application/json';
            }
            curl_setopt($handle, CURLOPT_POSTFIELDS, $payload);
        }
        curl_setopt($handle, CURLOPT_HTTPHEADER, $httpHeaders);
        curl_setopt($handle, CURLOPT_TIMEOUT, $this->config->getConnectTimeout());
        curl_setopt($handle, CURLOPT_LOW_SPEED_TIME, $this->config->getRequestTimeout());
        curl_setopt($handle, CURLOPT_URL, $url);

        /** @var string $response */
        $response = curl_exec($handle);

        $httpStatus = (int) curl_getinfo($handle, CURLINFO_HTTP_CODE);

        $curlErrorMessage = curl_error($handle);
        if (!empty($curlErrorMessage) && !($httpStatus >= 200 && $httpStatus <= 299)) {
            $curlErrorNumber = curl_errno($handle);
            trigger_error("[Ultimania] Error while communicating with Ultimania server. RequestId: $requestId; URL: $url; Parameters: ".print_r($payload, true)."; HTTP Status: $httpStatus; Curl error number: $curlErrorNumber; Message: $curlErrorMessage", E_USER_WARNING);
        }

        curl_close($handle);

        $jsonDecodedResponse = empty($response) ? null : json_decode($response, true);

        if ($jsonDecodedResponse !== null && !empty($jsonDecodedResponse['error'])) {
            trigger_error("[Ultimania] Received error from Ultimania server for RequestId $requestId: " . print_r($jsonDecodedResponse['error'], true), E_USER_WARNING);
        }

        return [
            'response' => $jsonDecodedResponse,
            'httpcode' => $httpStatus,
            'requestId' => $requestId
        ];
    }

    /**
     * @param UltimaniaPlayer $player
     * @param string $uid
     * @return string
     */
    public function getLinkForAdminRecManialink(UltimaniaPlayer $player, $uid) {
        return $this->adminRecManialinkUrl. '?uid=' . urlencode($uid) . '&login=' . urlencode($player->getLogin());
    }

    /**
     *
     * @param UltimaniaRecord $record
     * @return string
     */
    public function getLinkForViewReplayManialink($record) {
        return $this->viewReplayManialinkUrl . '?record_id=' . urlencode($record->getId()) . '&amp;cache_bust=' . uniqid();
    }

    /**
     * @return false|string
     */
    public function fetchNewestAvailableUltimaniaClientVersion() {
        $url = $this->ultimaniaApiServerUrl . '/current_version.txt';
        $versionRaw = file_get_contents($url);

        if ($versionRaw === false) {
            trigger_error('[Ultimania] Unable to get current version information from ' . $url, E_USER_WARNING);
            return false;
        }
        return trim($versionRaw);
    }

    /**
     * Fetch the Ultimania plugin PHP script from the server.
     * @param string $version
     * @return false|string
     */
    public function fetchUpdate($version) {
        return file_get_contents($this->ultimaniaApiServerUrl . '/versions_repository/' . $version . '.php_');
    }

    /**
     * @return false|string
     */
    public function fetchInfotextInMainWindow() {
        $url = $this->ultimaniaApiServerUrl . '/main_window_info_text.txt';
        $info = file_get_contents($url);

        if ($info === false) {
            trigger_error('[Ultimania] Unable to get window infobox content from ' . $url, E_USER_WARNING);
            return false;
        }
        return $info;
    }
}

class UltimaniaRecord {

    /** @var string|null */
    private $id;

    /** @var UltimaniaPlayer|null null if object constructed with login only */
    private $player;

    /** @var string|null null if object constructed with a full player object */
    private $playerLogin;

    /** @var string */
    private $mapUid;

    /** @var int */
    private $score;

    /** @var int|null  */
    private $addTime;

    /** @var bool|null */
    private $replayAvailable;

    /**
     * @param UltimaniaPlayer|string $player if string, only set the login
     * @param string $mapUid
     * @param int $score
     * @param int|null $addTime
     * @param string|null $id
     * @param boolean|null $replayAvailable
     */
    public function __construct($player, $mapUid, $score, $addTime = null, $id = null, $replayAvailable = null) {
        if (is_string($player)) {
            $this->playerLogin = $player;
        } else {
            $this->player = $player;
        }
        $this->mapUid = $mapUid;
        $this->score = $score;
        $this->addTime = $addTime;
        $this->id = $id;
        $this->replayAvailable = $replayAvailable;
    }

    /**
     * @return string|null
     */
    public function getId() {
        return $this->id;
    }

    /**
     * @param string|null $id
     * @return void
     */
    public function setId($id) {
        $this->id = $id;
    }

    /**
     * @return UltimaniaPlayer|null
     */
    public function getPlayer() {
        return $this->player;
    }

    /**
     * @return string|null
     */
    public function getPlayerLogin() {
        return $this->playerLogin;
    }

    /**
     * @return string
     */
    public function getMapUid() {
        return $this->mapUid;
    }

    /**
     * @return int
     */
    public function getScore() {
        return $this->score;
    }

    /**
     * @param int $score
     * @return void
     */
    public function setScore($score) {
        $this->score = $score;
    }

    /**
     * @return int|null
     */
    public function getAddTime() {
        return $this->addTime;
    }

    /**
     * @param int|null $addTime
     * @return void
     */
    public function setAddTime($addTime) {
        $this->addTime = $addTime;
    }

    /**
     * @return bool|null
     */
    public function isReplayAvailable() {
        return $this->replayAvailable;
    }

    /**
     * @param bool $replayAvailable
     * @return void
     */
    public function setReplayAvailable($replayAvailable) {
        $this->replayAvailable = $replayAvailable;
    }

}

class UltimaniaRecords {

    /** @var UltimaniaClient */
    private $ultiClient;

    /**
     * all records, always sorted in descending order
     * @var UltimaniaRecord[]
     */
    private $recordsOrderedByScore = [];

    /**
     * @param UltimaniaClient $ultiClient
     */
    public function __construct($ultiClient) {
        $this->ultiClient = $ultiClient;
    }

    /**
     * @return UltimaniaRecord[]
     */
    public function getAll() {
        return $this->recordsOrderedByScore;
    }

    /**
     * Return the best $limit records
     * @param int $limit
     * @return UltimaniaRecord[]
     */
    public function getLimitedBy($limit) {
        return array_slice($this->recordsOrderedByScore, 0, $limit);
    }

    /**
     * @param int $rank Starting at 1
     * @return UltimaniaRecord|null
     */
    public function getRecordByRank($rank) {
        if (isset($this->recordsOrderedByScore[$rank - 1])) {
            return $this->recordsOrderedByScore[$rank - 1];
        }
        return null;
    }

    /**
     * @param string $login
     * @return int|false
     */
    public function getRankOfPlayer($login) {
        $playerRecords = array_filter($this->recordsOrderedByScore, function($record) use ($login) {
            return $record->getPlayer()->getLogin() == $login;
        });

        if (empty($playerRecords)) {
            return false;
        }

        return array_keys($playerRecords)[0] + 1;
    }

    /**
     * @return bool
     */
    public function isEmpty() {
        return empty($this->recordsOrderedByScore);
    }

    /**
     * @param string $forMapUid
     * @return void
     */
    public function refresh($forMapUid) {
        $this->recordsOrderedByScore = $this->ultiClient->getRecords($forMapUid);
    }

    /**
     * @param UltimaniaRecord $newRecord
     * @param string|null $replayContent Replay, if available
     * @return UltimaniaRecordImprovement|null
     */
    public function saveRecord($newRecord, $replayContent) {
        $improvement = $this->localInsertOrUpdate($newRecord);

        if ($improvement->getType() != UltimaniaRecordImprovement::TYPE_NO_IMPROVEMENT) {
            $savedRecord = $this->ultiClient->submitRecord($newRecord, $newRecord->getMapUid());
            if ($savedRecord == null) {
                return null;
            }

            if ($replayContent !== null) {
                $replayAvailable = $this->ultiClient->submitReplay($savedRecord->getId(), $replayContent)['replay_available'];
            } else {
                $replayAvailable = false;
            }

            // post-saving updates
            $referenceToLocallySavedRecord = $this->getRecordByLogin($newRecord->getPlayer()->getLogin());
            $referenceToLocallySavedRecord->setId($savedRecord->getId());
            $referenceToLocallySavedRecord->setReplayAvailable($replayAvailable);
        }

        return $improvement;
    }

    /**
     * @return array{string: UltimaniaRecord} {login: UltimaniaRecord}[]
     */
    public function getRecordsIndexedByLogin() {
        $indexedByLogin = [];
        foreach ($this->recordsOrderedByScore as $record) {
            $indexedByLogin[$record->getPlayer()->getLogin()] = $record;
        }
        return $indexedByLogin; /* @phpstan-ignore-line It doesn't really get the return type is actually correct */
    }

    /**
     * @param string $login
     * @return UltimaniaRecord|null
     */
    public function getRecordByLogin($login) {
        $records = $this->getRecordsIndexedByLogin();
        if (isset($records[$login])) {
            return $records[$login];
        }
        return null;
    }

    /**
     * @param UltimaniaRecord $newRecord
     * @return UltimaniaRecordImprovement
     */
    private function localInsertOrUpdate($newRecord) {
        $improvement = new UltimaniaRecordImprovement();

        $referenceToPreviousRecord = $this->getRecordByLogin($newRecord->getPlayer()->getLogin());

        $improvement->setPreviousRecord($this->cloneIfIsObject($referenceToPreviousRecord));
        $improvement->setPreviousRank($this->getRankByLogin($newRecord->getPlayer()->getLogin()));

        $this->updateScoreOfRecordIfImprovedOrInsert($referenceToPreviousRecord, $newRecord);

        $improvement->setNewRecord($newRecord);
        $improvement->setNewRank($this->getRankByLogin($newRecord->getPlayer()->getLogin()));

        return $improvement;
    }

    /**
     * @param UltimaniaRecord|null $previousRecord
     * @param UltimaniaRecord $newRecord
     * @return void
     */
    private function updateScoreOfRecordIfImprovedOrInsert($previousRecord, UltimaniaRecord $newRecord) {
        if ($previousRecord instanceof UltimaniaRecord) {
            if ($newRecord->getScore() > $previousRecord->getScore()) {
                $previousRecord->setScore($newRecord->getScore());
            }
        } else {
            $this->recordsOrderedByScore[] = $newRecord;
        }
        usort($this->recordsOrderedByScore, "self::sortRecordsDescCallback"); /* @phpstan-ignore-line PHPStand doesn't understand this string is a callback. */
    }

    /**
     * usort callback for records.
     * Use with usort($array, 'ulti_sortRecordsDesc')
     * @param UltimaniaRecord $a
     * @param UltimaniaRecord $b
     * @return int
     */
    function sortRecordsDescCallback(UltimaniaRecord $a, UltimaniaRecord $b) {
        if ($a->getScore() == $b->getScore()) {
            return 0;
        }

        return ($a->getScore() > $b->getScore()) ? -1 : 1;
    }

    /**
     * Returns the rank of a player or -1 if he doesn't have a record yet
     * @param string $login
     * @return int
     */
    private function getRankByLogin($login) {
        foreach ($this->getAll() as $i => $record) {
            if ($record->getPlayer()->getLogin() === $login) {
                return $i + 1;
            }
        }
        return -1;
    }

    /**
     * @template T
     * @param T $obj
     * @return T|null
     */
    private function cloneIfIsObject($obj) {
        if (is_object($obj)) {
            return clone $obj;
        }
        return null;
    }
}

class UltimaniaRecordImprovement {
    const TYPE_NO_IMPROVEMENT = 'NO_IMPROVEMENT';
    const TYPE_NEW = 'NEW'; // secured
    const TYPE_EQUAL = 'EQUAL';
    const TYPE_NEW_RANK = 'NEW_RANK'; // gained
    const TYPE_FIRST = 'FIRST';
    const TYPE_UNKNOWN = 'UNKNOWN';

    /** @var int */
    private $previousRank;
    /** @var null|UltimaniaRecord */
    private $previousRecord;
    /** @var int */
    private $newRank;
    /** @var UltimaniaRecord */
    private $newRecord;

    /**
     * @return string
     */
    public function getType() {
        // special case where null can occur.
        // do this first to preven nullpointer exceptions in the other checks
        if ($this->hasPreviousRecord()) {
            return self::TYPE_FIRST;
        }

        if ($this->newRecord->getScore() < $this->previousRecord->getScore()) {
            return self::TYPE_NO_IMPROVEMENT;
        }
        if ($this->newRecord->getScore() > $this->previousRecord->getScore() &&
            $this->newRank == $this->previousRank) {
            return self::TYPE_NEW;
        }
        if ($this->newRecord->getScore() == $this->previousRecord->getScore()) {
            return self::TYPE_EQUAL;
        }
        if ($this->newRank < $this->previousRank) {
            return self::TYPE_NEW_RANK;
        }
        return self::TYPE_UNKNOWN;
    }

    /**
     * @return int Difference between new and previous score. Positive if improved.
     */
    public function getRecordDifference() {
        return $this->getNewRecord()->getScore() - $this->getPreviousRecord()->getScore();
    }

    /**
     * @return int
     */
    public function getPreviousRank() {
        return $this->previousRank;
    }

    /**
     * @param int $previousRank
     * @return void
     */
    public function setPreviousRank($previousRank) {
        $this->previousRank = $previousRank;
    }

    /**
     * @return UltimaniaRecord|null
     */
    public function getPreviousRecord() {
        return $this->previousRecord;
    }

    /**
     * @param UltimaniaRecord|null $previousRecord
     * @return void
     */
    public function setPreviousRecord($previousRecord) {
        $this->previousRecord = $previousRecord;
    }

    /**
     * @return int
     */
    public function getNewRank() {
        return $this->newRank;
    }

    /**
     * @param int $newRank
     * @return void
     */
    public function setNewRank($newRank) {
        $this->newRank = $newRank;
    }

    /**
     * @return UltimaniaRecord
     */
    public function getNewRecord() {
        return $this->newRecord;
    }

    /**
     * @param UltimaniaRecord $newRecord
     * @return void
     */
    public function setNewRecord($newRecord) {
        $this->newRecord = $newRecord;
    }

    /**
     * @return bool
     */
    private function hasPreviousRecord() {
        return empty($this->previousRecord);
    }
}

class UltimaniaPlayer {

    /** @var string */
    private $login;

    /** @var string */
    private $nick;

    /** @var bool|null */
    private $allowReplayDownload;

    /** @var bool|null */
    private $banned;

    /**
     * @param string $login
     * @param string $nick
     * @param bool|null $allowReplayDownload
     * @param bool|null $banned
     */
    public function __construct($login, $nick, $allowReplayDownload = null, $banned = null) {
        $this->login = $login;
        $this->nick = $nick;
        $this->allowReplayDownload = $allowReplayDownload;
        $this->banned = $banned;
    }

    /**
     * @return string
     */
    public function getLogin() {
        return $this->login;
    }

    /**
     * @return string
     */
    public function getNick() {
        return $this->nick;
    }

    /**
     * @return bool|null
     */
    public function isAllowReplayDownload() {
        return $this->allowReplayDownload;
    }

    /**
     * @param bool $allowReplayDownload
     * @return void
     */
    public function setAllowReplayDownload($allowReplayDownload) {
        $this->allowReplayDownload = $allowReplayDownload;
    }

    /**
     * @return bool|null
     */
    public function isBanned() {
        return $this->banned;
    }
}

class UltimaniaLegacyAdapter {
    /** @var Aseco */
    private $aseco;
    /** @var UltimaniaConfig */
    private $ultiConfig;
    /** @var UltimaniaRecords */
    private $ultiRecords;

    /**
     * @param UltimaniaConfig $ultiConfig
     * @param UltimaniaRecords $ultiRecords
     */
    public function __construct($ultiConfig, $ultiRecords) {
        global $aseco;
        $this->aseco = $aseco;
        $this->ultiConfig = $ultiConfig;
        $this->ultiRecords = $ultiRecords;
    }

    /**
     * @param string $name
     * @return array{'login': string, "nick": string, "score": int}[]|void
     */
    public function __get($name) {
        if ($name === 'records') {
            $ultiRecords = $this->ultiRecords->getLimitedBy($this->ultiConfig->getNumberOfRecordsDisplayLimit());
            return $this->mapUltimaniaRecordsToLegacyArray($ultiRecords);
        }
    }

    /**
     * @param UltimaniaRecord[] $records
     * @return void
     */
    public function mapRecordsLoadedEventToLegacyEvent($records) {
        $this->aseco->releaseEvent(ULTI_EVENT_RECORDS_LOADED_LEGACY, $this->mapUltimaniaRecordsToLegacyArray($records));
    }

    /**
     * @param UltimaniaRecord $record
     * @return void
     */
    public function mapRecordEventToLegacyEvent($record) {
        $this->aseco->releaseEvent(ULTI_EVENT_RECORD_LEGACY, $this->mapUltimaniaRecordToLegacyArray($record));
    }

    /**
     * @param UltimaniaRecord $record
     * @return array{'login': string, "nick": string, "score": int}
     */
    private function mapUltimaniaRecordToLegacyArray(UltimaniaRecord $record) {
        return [
            'login' => $record->getPlayer()->getLogin(),
            'nick' => $record->getPlayer()->getNick(),
            'score' => $record->getScore(),
        ];
    }

    /**
     * @param UltimaniaRecord[] $ultiRecords
     * @return array{'login': string, "nick": string, "score": int}[]
     */
    private function mapUltimaniaRecordsToLegacyArray($ultiRecords) {
        return array_map('self::mapUltimaniaRecordToLegacyArray', $ultiRecords); /* @phpstan-ignore-line PHPStand doesn't understand this string is a callback. */
    }
}

class UltimaniaXasecoAdapter {

    /** @var Aseco */
    private $aseco;

    public function __construct() {
        global $aseco;

        $this->aseco = $aseco;
    }

    /**
     * In exceptional cases, direct acces to Aseco is required.
     * Avoid using this method!
     * @return Aseco
     */
    public function getAsecoObject() {
        return $this->aseco;
    }

    /**
     * @param string $message
     * @return void
     */
    public function console($message) {
        $this->aseco->console($message);
    }

    /**
     * @return bool
     */
    public function isXasecoStartingUp() {
        return $this->aseco->startup_phase;
    }

    /**
     * @param String $message
     * @return void
     */
    public function chatSendServerMessage($message) {
        $this->aseco->client->query('ChatSendServerMessage', $message);
    }

    /**
     * @param String $message
     * @param Player $player
     * @return void
     */
    public function chatSendServerMessageToPlayer($message, Player $player) {
        $this->aseco->client->query('ChatSendServerMessageToLogin', $message, $player->login);
    }

    /**
     * @param Player $player
     * @return bool
     */
    public function isMasterAdmin(Player $player) {
        return $this->aseco->isMasterAdmin($player);
    }

    /**
     * @return Challenge
     */
    public function getCurrentChallengeObject() {
        return $this->aseco->server->challenge;
    }

    /**
     * @return string
     */
    public function getGameType() {
        return $this->aseco->server->getGame();
    }

    /**
     * @return Gameinfo
     */
    public function getGameInfo() {
        return $this->aseco->server->gameinfo;
    }

    /**
     * @return Server
     */
    public function getServerObject() {
        return $this->aseco->server;
    }


    /**
     * @return Player[]
     */
    public function getPlayerList() {
        return $this->aseco->server->players->player_list;
    }

    /**
     * @param string $login
     * @return Player
     */
    public function getPlayerObjectFromLogin($login) {
        return $this->getPlayerList()[$login];
    }

    /**
     * @param int $playerUid server's player uid
     * @param Player $player
     * @param int $action action id
     * @return void
     */
    public function releaseOnPlayerManialinkPageAnswerEvent($playerUid, Player $player, $action) {
        $this->aseco->releaseEvent('onPlayerManialinkPageAnswer',  array($playerUid, $player->login, $action));
    }

    /**
     * @param string $name
     * @param mixed $content
     * @return void
     */
    public function releaseEvent($name, $content) {
        $this->aseco->releaseEvent($name, $content);
    }

    /**
     * @param string $text
     * @return string
     */
    public function formatColors($text) {
        return $this->aseco->formatColors($text);
    }

    /**
     * @param Player $player
     * @param string $xml Manialink XML
     * @param int $autohideTimeout timeout in seconds
     * @param bool $hideOnClick
     * @return void
     */
    public function sendManialinkToPlayer($player, $xml, $autohideTimeout = 0, $hideOnClick = false) {
        $this->aseco->client->addCall('SendDisplayManialinkPageToLogin', array($player->login, $xml, $autohideTimeout, $hideOnClick));
    }

    /**
     * @param string $xml Manialink XML
     * @param int $autohideTimeout timeout in seconds
     * @param bool $hideOnClick
     * @return void
     */
    public function sendManialinkToEveryone($xml, $autohideTimeout = 0, $hideOnClick = false) {
        $this->aseco->client->addCall('SendDisplayManialinkPage', array($xml, $autohideTimeout, $hideOnClick));
    }

    /**
     * @param string $plugin filename of the plugin?
     * @param string $author
     * @param string $version
     * @return void
     */
    public function registerWithThirdpartyPluginsUpToDate($plugin, $author, $version) {
        $this->aseco->plugin_versions[] = array( /** @phpstan-ignore-line */
            'plugin' => $plugin,
            'author' => $author,
            'version' => $version
        );
    }

    /**
     * @param Player $player
     * @return string|null Replay file bytes or null on failure
     */
    public function getBestReplayForPlayer(Player $player) {
        $replayFilename = 'ulti_replay_temp_file__delete_me.Replay.Gbx';
        $this->aseco->client->query('SaveBestGhostsReplay',
            $player->login,
            $replayFilename
        );
        $response = $this->aseco->client->getResponse();

        if ($response != "1") {
            trigger_error('[Ultimania] Server is unable to save replay into "GameData/Tracks/Replays/". Please make sure the server has write access in this folder.', E_USER_WARNING);
            $this->console('Server error message: ' . $this->aseco->client->getErrorMessage());
            return null;
        }

        $replayFileName = rtrim(rtrim($this->aseco->server->gamedir, '/'), '\\') . '/Tracks/Replays/' . $replayFilename;
        $replayFile = file_get_contents($replayFileName);

        if ($replayFile === false) {
            trigger_error('[Ultimania] Unable to read previously saved replay from ' . $replayFileName, E_USER_WARNING);
            return null;
        }
        unlink($replayFileName);

        return $replayFile;
    }
}

/*
 * @ plugin authors: don't access these objects directly.
 * Use the events onUltimaniaRecordsLoadedApi2 and onUltimaniaRecordApi2 instead.
 * I don't guarantee not to break the interface.
 */

$ultiConfig = UltimaniaConfig::instantiateFromFile('ultimania.xml');

$ultiDtoMapper = new UltimaniaDtoMapper();
$ultiClient = new UltimaniaClient($ultiConfig, $ultiDtoMapper);
$ultiRecords = new UltimaniaRecords($ultiClient);
$ultiXasecoAdapter = new UltimaniaXasecoAdapter();
$ultiMainClass = new Ultimania($ultiConfig, $ultiRecords, $ultiXasecoAdapter, $ultiClient);

// deprecated! Use onUltimaniaRecordsLoadedApi2 and onUltimaniaRecordApi2 instead
$ulti = new UltimaniaLegacyAdapter($ultiConfig, $ultiRecords);

/**
 * Chatcommands
 */

function chat_ultirankinfo($aseco, $command) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onChatUltiRankInfo($command['author'], $command['params']);
}

function chat_ultiwindow($aseco, $command) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onChatUltiWindow($command['author']);
}

function chat_ultiupdate($aseco, $command) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onChatUltiUpdate($command['author']);
}

function chat_ultilist($aseco, $command) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onChatUltiList($command['author']);
}

/**
 * Callbacks
 */
function ulti_onSync($aseco) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onStartup();
}
function ulti_onNewChallenge2($aseco, $challenge_item) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onNewChallenge($challenge_item);
}
function ulti_onPlayerFinish($aseco, $finish) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onPlayerFinish($finish);
}
function ulti_onPlayerConnect($aseco, $player) { /** @phpstan-ignore-line */
    global $ultiMainClass, $ultiXasecoAdapter;

    if (! $ultiXasecoAdapter->isXasecoStartingUp()) {
        $ultiMainClass->onPlayerConnect($player);
    }
}
function ulti_onEndRace1($aseco, $race) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onEndRace();
}
function ulti_onEverySecond($aseco) { /** @phpstan-ignore-line */
    global $ultiMainClass;
    $ultiMainClass->onEverySecond();
}
function ulti_onMlAnswer($aseco, $answer) { /** @phpstan-ignore-line */
    global $ultiMainClass, $ultiXasecoAdapter;
    $ultiMainClass->onMlAnswer($answer[0], $ultiXasecoAdapter->getPlayerObjectFromLogin($answer[1]), $answer[2]);
}
function ulti_onMenuLoaded($aseco, $menu) { /** @phpstan-ignore-line */
    // Doc: http://fish.stabb.de/index.php5?page=downloads&subpage=148
    $menu->addEntry('tracks', 'fulllist', true, 'Ultimania', 'ultifolder', '');

    $menu->addEntry('', 'tracks', false, 'Ultimania', 'ulti');
    $menu->addEntry('ulti', '', true, 'Full recordlist', 'ulti_fulllist', '/ultilist');
    $menu->addEntry('ulti', '', true, 'Open window', 'ulti_openwindow', '/ultiwindow');
    $menu->addEntry('ulti', '', true, '$ff0Update plugin', 'ulti_openwindow', '/ultiupdate', 'MasterAdmin');
}
function ulti_onUltimaniaRecordsLoadedApi2($aseco, $records) { /** @phpstan-ignore-line */
    global $ulti;
    $ulti->mapRecordsLoadedEventToLegacyEvent($records);
}
function ulti_onUltimaniaRecordApi2($aseco, $record) { /** @phpstan-ignore-line */
    global $ulti;
    $ulti->mapRecordEventToLegacyEvent($record);
}
